@@ -1,5 +1,4 @@
 import argparse
-import calendar
 import concurrent.futures
 import datetime as dt
 import gzip
@@ -8,87 +7,39 @@ import os
 import shutil
 import sys
 import time
+from datetime import datetime
 
 import dataframe_image as dfi
+import geopandas as gpd
 import osmium
 import pandas as pd
-import pytz
 import requests
 from osmium.replication.server import ReplicationServer
+from shapely.geometry import box
 
 from osmsg.utils import verify_me_osm
 
+from .changefiles import (
+    get_prev_hour,
+    get_prev_year_dates,
+    in_local_timezone,
+    previous_day,
+    previous_month,
+    previous_week,
+    seq_to_timestamp,
+    strip_utc,
+)
+from .changesets import ChangesetToolKit
+
 users_temp = {}
 users = {}
+hashtag_changesets = {}
 
-from datetime import datetime, timedelta
+# read the GeoJSON file
+countries_df = gpd.read_file("data/countries_un.geojson")
 
 
-def in_local_timezone(date, timezone):
-    tz = dt.timezone.utc
-    if timezone == "Nepal":
-        # Set the timezone to Nepal
-        tz = pytz.timezone("Asia/Kathmandu")
-
-    return date.astimezone(tz)
-
-
-def strip_utc(date_str, timezone):
-    tz = dt.timezone.utc
-    given_dt = dt.datetime.strptime(date_str, "%Y-%m-%d").replace(tzinfo=tz)
-
-    if timezone == "Nepal":
-        # Set the timezone to Nepal
-        tz = pytz.timezone("Asia/Kathmandu")
-
-    return given_dt.astimezone(tz)
-
-
-def get_prev_year_dates(timezone):
-    # Get today's date
-    today = datetime.utcnow()
-
-    # Get the start and end date of the previous year
-    prev_year_start = datetime(today.year - 1, 1, 1)
-    prev_year_end = datetime(today.year - 1, 12, 31)
-
-    return strip_utc(prev_year_start.strftime("%Y-%m-%d"), timezone), strip_utc(
-        prev_year_end.strftime("%Y-%m-%d"), timezone
-    )
-
-
-def previous_month(timezone):
-    today = datetime.utcnow()
-    month = today.month
-    year = today.year
-    if today.month == 1:
-        month = 13
-        year -= 1
-    prev_month_start = datetime(year, month - 1, 1)
-    prev_month_end = datetime(year, month - 1, calendar.monthrange(year, month - 1)[1])
-    return strip_utc(prev_month_start.strftime("%Y-%m-%d"), timezone), strip_utc(
-        prev_month_end.strftime("%Y-%m-%d"), timezone
-    )
-
-
-def previous_day(timezone):
-    today = datetime.today()
-    previous_day = today - timedelta(days=1)
-    return strip_utc(previous_day.strftime("%Y-%m-%d"), timezone), strip_utc(
-        today.strftime("%Y-%m-%d"), timezone
-    )
-
-
-def previous_week(timezone):
-    today = datetime.today()
-    start_date = today - timedelta(days=today.weekday() + 7)
-    end_date = start_date + timedelta(days=6)
-    return strip_utc(start_date.strftime("%Y-%m-%d"), timezone), strip_utc(
-        end_date.strftime("%Y-%m-%d"), timezone
-    )
-
-
-def calculate_stats(user, uname, changeset, version, tags, osm_type):
+def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
     tags_to_collect = list(additional_tags) if additional_tags else None
     if version == 1:
         action = "create"
@@ -102,7 +53,20 @@ def calculate_stats(user, uname, changeset, version, tags, osm_type):
         if changeset not in users_temp[user]["changesets"]:
             users_temp[user]["changesets"].append(changeset)
         users[user]["changesets"] = len(users_temp[user]["changesets"])
+        if hashtags:
+            for ch in hashtag_changesets[changeset]:
+                if ch not in users[user]["countries"]:
+                    users[user]["countries"].append(ch)
         users[user][osm_type][action] += 1
+        if osm_type == "nodes" and tags:
+            users[user]["poi"][action] += 1
+        if wild_tags:
+            for key, value in tags:
+                if key in users[user][f"tags_{action}"]:
+                    users[user][f"tags_{action}"][key] += 1
+                else:
+                    users[user][f"tags_{action}"][key] = 1
+
         if tags_to_collect:
             for tag in tags_to_collect:
                 if tag in tags:
@@ -116,7 +80,10 @@ def calculate_stats(user, uname, changeset, version, tags, osm_type):
             "nodes": {"create": 0, "modify": 0, "delete": 0},
             "ways": {"create": 0, "modify": 0, "delete": 0},
             "relations": {"create": 0, "modify": 0, "delete": 0},
+            "poi": {"create": 0, "modify": 0, "delete": 0},  # nodes that has tags
         }
+        if hashtags:
+            users[user]["countries"] = hashtag_changesets[changeset]
         if tags_to_collect:
             for tag in tags_to_collect:
                 users[user][tag] = {"create": 0, "modify": 0, "delete": 0}
@@ -125,23 +92,65 @@ def calculate_stats(user, uname, changeset, version, tags, osm_type):
             users_temp[user]["changesets"].append(changeset)
         users[user]["changesets"] = len(users_temp[user]["changesets"])
         users[user][osm_type][action] = 1
+        if wild_tags:
+            users[user]["tags_create"] = {}
+            users[user]["tags_modify"] = {}
+            users[user]["tags_delete"] = {}
+
+            for tag, value in tags:
+                users[user][f"tags_{action}"][tag] = 1
         if tags_to_collect:
             for tag in tags_to_collect:
                 if tag in tags:
                     users[user][tag][action] = 1
 
 
+def calculate_stats(user, uname, changeset, version, tags, osm_type):
+    if hashtags:  # intersect with changesets
+        if (
+            len(hashtag_changesets) > 0
+        ):  # make sure there are changesets to intersect if not meaning hashtag changeset not found no need to go for changefiles
+            if changeset in hashtag_changesets.keys():
+                collect_changefile_stats(
+                    user, uname, changeset, version, tags, osm_type
+                )
+    else:  # collect everything
+        collect_changefile_stats(user, uname, changeset, version, tags, osm_type)
+
+
+class ChangesetHandler(osmium.SimpleHandler):
+    def __init__(self):
+        super(ChangesetHandler, self).__init__()
+
+    def changeset(self, c):
+        if "comment" in c.tags:
+            if any(elem.lower() in c.tags["comment"].lower() for elem in hashtags):
+                hashtag_changesets[c.id] = []
+                # get bbox
+                bounds = str(c.bounds)
+                bbox_list = bounds.strip("()").split(" ")
+                minx, miny = bbox_list[0].split("/")
+                maxx, maxy = bbox_list[1].split("/")
+                bbox = box(float(minx), float(miny), float(maxx), float(maxy))
+                # Create a point for the centroid of the bounding box
+                centroid = bbox.centroid
+                intersected_rows = countries_df[countries_df.intersects(centroid)]
+                for i, row in intersected_rows.iterrows():
+                    if row["name"] not in hashtag_changesets[c.id]:
+                        hashtag_changesets[c.id].append(row["name"])
+
+
 class ChangefileHandler(osmium.SimpleHandler):
     def __init__(self):
         super(ChangefileHandler, self).__init__()
 
     def node(self, n):
         if n.timestamp >= start_date_utc and n.timestamp <= end_date_utc:
-
             calculate_stats(n.uid, n.user, n.changeset, n.version, n.tags, "nodes")
 
     def way(self, w):
         if w.timestamp >= start_date_utc and w.timestamp <= end_date_utc:
+
             calculate_stats(w.uid, w.user, w.changeset, w.version, w.tags, "ways")
 
     def relation(self, r):
@@ -165,7 +174,7 @@ def process_changefiles(url):
     if not os.path.exists(file_path):
         # Read the cookies from the file
 
-        if "geofabrik" in url:
+        if "geofabrik" in url.lower():
             cookies_fmt = {}
             test = cookies.split("=")
             # name, value = line.strip().split("=")
@@ -183,21 +192,61 @@ def process_changefiles(url):
             f.write(file_data)
 
     # Open the .osc.gz file in read-only mode
-    print(file_path)
     handler = ChangefileHandler()
     with gzip.open(file_path, "rb") as f_in, open(file_path[:-3], "wb") as f_out:
         shutil.copyfileobj(f_in, f_out)
 
     handler.apply_file(file_path[:-3])
+    print(f"Finished {url}")
+
+
+def process_changesets(url):
+    print(f"Processing {url}")
+    url_splitted_list = url.split("/")
+    temp_path = os.path.join(os.getcwd(), "temp/changesets")
+    if not os.path.exists(temp_path):
+        os.makedirs(temp_path)
+    file_path = os.path.join(
+        temp_path,
+        f"{url_splitted_list[-3]}_{url_splitted_list[-2]}_{url_splitted_list[-1]}",
+    )
+    if not os.path.exists(file_path):
+
+        response = requests.get(url)
+        if not response.status_code == 200:
+            sys.exit()
+
+        file_data = response.content
+
+        with open(file_path, "wb") as f:
+            f.write(file_data)
+
+    # Open the .osc.gz file in read-only mode
+
+    handler = ChangesetHandler()
+    with gzip.open(file_path, "rb") as f_in, open(file_path[:-3], "wb") as f_out:
+        shutil.copyfileobj(f_in, f_out)
+
+    handler.apply_file(file_path[:-3])
 
 
-def get_download_urls_changefiles(start_date, end_date, base_url, timezone):
+def get_download_urls_changefiles(
+    start_date, end_date, base_url, timezone, scan_early_seq=False
+):
     repl = ReplicationServer(base_url)
 
     # Gets sequence id using timestamp we get from osm api using pyosmium tool
     seq = repl.timestamp_to_sequence(start_date)
-    # going one step back to cover all changes
-    seq = seq - 1
+    # going one step back to cover all changes only if it is not already behind
+    if (
+        scan_early_seq
+        or (start_date - seq_to_timestamp(repl.get_state_url(seq), timezone)).days < 1
+    ):
+        print(
+            f"Initial fetched difference is {(seq_to_timestamp(repl.get_state_url(seq), timezone) - start_date).days} Hence , Reducing Sequence to Cover all Changes"
+        )
+        seq = seq - 1
+
     start_seq = seq
     start_seq_url = repl.get_state_url(start_seq)
 
@@ -230,7 +279,7 @@ def get_download_urls_changefiles(start_date, end_date, base_url, timezone):
             last_ts = end_date
 
     print(
-        f"You have supplied {start_date} : {seq} to {last_ts} : {last_seq} . Latest Server Fetched is : {server_seq} & {in_local_timezone(server_ts,timezone)} on {base_url}"
+        f"You have supplied {start_date} : {seq} to {last_ts} : {last_seq} . Latest Server Fetched is : {server_seq} & {in_local_timezone(server_ts,timezone)} on {base_url}\nGoing from {seq_to_timestamp(repl.get_state_url(seq), timezone)} to {seq_to_timestamp(repl.get_state_url(last_seq), timezone)}"
     )
 
     if seq >= last_seq:
@@ -252,27 +301,6 @@ def get_download_urls_changefiles(start_date, end_date, base_url, timezone):
     return download_urls, server_ts, start_seq, last_seq, start_seq_url, end_seq_url
 
 
-def seq_to_timestamp(url, timezone):
-    response = requests.get(url)
-    rtxt = response.text
-    # find the index of the "timestamp=" substring
-    timestamp_start = rtxt.find("timestamp=") + len("timestamp=")
-
-    # find the index of the next newline character
-    timestamp_end = rtxt.find("\n", timestamp_start)
-
-    # extract the substring between the two indexes
-    timestamp = rtxt[timestamp_start:timestamp_end]
-    timestamp_obj = dt.datetime.strptime(timestamp, "%Y-%m-%dT%H\:%M\:%SZ")
-    timestamp_obj = timestamp_obj.replace(tzinfo=dt.timezone.utc)
-    tz = dt.timezone.utc
-    if timezone == "Nepal":
-        # Set the timezone to Nepal
-        tz = pytz.timezone("Asia/Kathmandu")
-
-    return timestamp_obj.astimezone(tz)
-
-
 def auth(username, password):
     print("Authenticating...")
     try:
@@ -288,9 +316,15 @@ def main():
     parser = argparse.ArgumentParser()
     parser.add_argument("--start_date", help="Start date in the format YYYY-MM-DD")
     parser.add_argument("--end_date", help="End date in the format YYYY-MM-DD")
-    parser.add_argument("--username", required=True, help="Your OSM Username")
     parser.add_argument(
-        "--password", required=True, help="Your OSM Password", default=argparse.SUPPRESS
+        "--username",
+        default=None,
+        help="Your OSM Username : Only required for Geofabrik Internal Changefiles",
+    )
+    parser.add_argument(
+        "--password",
+        default=None,
+        help="Your OSM Password : Only required for Geofabrik Internal Changefiles",
     )
     parser.add_argument(
         "--timezone",
@@ -311,6 +345,19 @@ def main():
         help="Additional stats to collect : List of tags key",
     )
 
+    parser.add_argument(
+        "--hashtags",
+        nargs="+",
+        type=str,
+        help="Hashtags Statstics to Collect : List of hashtags , Limited until daily stats for now",
+    )
+    parser.add_argument(
+        "--force",
+        action="store_true",
+        help="Force for the Hashtag Replication fetch if it is greater than a day interval",
+        default=False,
+    )
+
     parser.add_argument(
         "--rows",
         type=int,
@@ -324,6 +371,14 @@ def main():
     parser.add_argument("--extract_last_day", action="store_true", default=False)
     parser.add_argument("--extract_last_month", action="store_true", default=False)
     parser.add_argument("--extract_last_year", action="store_true", default=False)
+    parser.add_argument("--extract_last_hour", action="store_true", default=False)
+
+    parser.add_argument(
+        "--wild_tags",
+        action="store_true",
+        help="Extract statistics of all of the unique tags and its count",
+        default=False,
+    )
 
     parser.add_argument(
         "--exclude_date_in_name",
@@ -341,41 +396,58 @@ def main():
     )
     parser.add_argument(
         "--read_from_metadata",
-        help="Location of metadata to pick start date from previous run's end_date",
+        help="Location of metadata to pick start date from previous run's end_date , Generally used if you want to run bot on regular interval using cron/service",
     )
 
     args = parser.parse_args()
     if args.start_date:
-        start_date = dt.datetime.strptime(args.start_date, "%Y-%m-%d").replace(
-            tzinfo=dt.timezone.utc
+        start_date = strip_utc(
+            dt.datetime.strptime(args.start_date, "%Y-%m-%d %H:%M:%S%z"), args.timezone
         )
 
-    if not args.start_date:
+    if not args.start_date or not args.end_date:
         if (
             args.extract_last_week
             or args.extract_last_day
             or args.extract_last_month
             or args.extract_last_year
+            or args.extract_last_hour
         ):
             pass
         else:
-            print("Supply start_date")
+            print("ERR: Supply start_date & End_date")
             sys.exit()
 
     if args.end_date:
-        end_date = dt.datetime.strptime(args.end_date, "%Y-%m-%d").replace(
-            tzinfo=dt.timezone.utc
+        end_date = strip_utc(
+            dt.datetime.strptime(args.end_date, "%Y-%m-%d %H:%M:%S%z"), args.timezone
         )
     start_time = time.time()
 
     global additional_tags
     global cookies
+    global wild_tags
+    global hashtags
+    wild_tags = args.wild_tags
     additional_tags = args.tags
+    hashtags = args.hashtags
     cookies = None
-    if "geofabrik" in args.url:
+    if "geofabrik" in args.url.lower():
+        if args.username is None:
+            args.username = os.environ.get("OSM_USERNAME")
+        if args.password is None:
+            args.password = os.environ.get("OSM_PASSWORD")
+
+        if not (args.username and args.password):
+            assert (
+                args.username and args.password
+            ), "OSM username and password are required for geofabrik url"
         cookies = auth(args.username, args.password)
     print("Script Started")
 
+    if args.extract_last_hour:
+        start_date, end_date = get_prev_hour(args.timezone)
+
     if args.extract_last_year:
         start_date, end_date = get_prev_year_dates(args.timezone)
 
@@ -406,7 +478,34 @@ def main():
         print("Err: Start date and end date are equal")
         sys.exit()
 
-    print("Generating Download Urls")
+    if (end_date - start_date).days > 1:
+        if args.hashtags:
+            print(
+                "Replication for Changeset is minutely , To process more than a day data it might take a while"
+            )
+            if not args.force:
+                sys.exit()
+
+    if args.hashtags:
+
+        Changeset = ChangesetToolKit()
+        (
+            changeset_download_urls,
+            changeset_start_seq,
+            changeset_end_seq,
+        ) = Changeset.get_download_urls(start_date, end_date)
+        print(
+            f"You have supplied start_date as : {start_date} and end_date as : {end_date} , Processing Changeset from {strip_utc(Changeset.sequence_to_timestamp(changeset_start_seq),args.timezone)} to {strip_utc(Changeset.sequence_to_timestamp(changeset_end_seq),args.timezone)}"
+        )
+        with concurrent.futures.ThreadPoolExecutor() as executor:
+            # Use `map` to apply the `download_image` function to each element in the `urls` list
+            executor.map(process_changesets, changeset_download_urls)
+        print("Changeset Processing Finished")
+        end_date = strip_utc(
+            Changeset.sequence_to_timestamp(changeset_end_seq), args.timezone
+        )
+
+    print("Changefiles : Generating Download Urls")
     (
         download_urls,
         server_ts,
@@ -421,16 +520,13 @@ def main():
         )
         end_date = server_ts
         if start_date >= server_ts:
-            print("Data is not available after start date ")
+            print("Err: Data is not available after start date ")
             sys.exit()
     global end_date_utc
     global start_date_utc
 
     start_date_utc = start_date
     end_date_utc = end_date
-    print("Download urls Generated")
-
-    print("Starting Thread Processing")
     temp_path = os.path.join(os.getcwd(), "temp/changefiles")
     if not os.path.exists(temp_path):
         os.makedirs(temp_path)
@@ -439,11 +535,45 @@ def main():
     with concurrent.futures.ThreadPoolExecutor() as executor:
         # Use `map` to apply the `download_image` function to each element in the `urls` list
         executor.map(process_changefiles, download_urls)
+    print("Changefiles Processing Finished")
     os.chdir(os.getcwd())
     shutil.rmtree("temp")
     if len(users) > 1:
-
+        # print(users)
+        if args.wild_tags:
+            for user in users:
+                users[user]["tags_create"] = json.dumps(
+                    dict(
+                        sorted(
+                            users[user]["tags_create"].items(),
+                            key=lambda item: item[1],
+                            reverse=True,
+                        )
+                    )
+                )
+                users[user]["tags_modify"] = json.dumps(
+                    dict(
+                        sorted(
+                            users[user]["tags_modify"].items(),
+                            key=lambda item: item[1],
+                            reverse=True,
+                        )
+                    )
+                )
+                users[user]["tags_delete"] = json.dumps(
+                    dict(
+                        sorted(
+                            users[user]["tags_delete"].items(),
+                            key=lambda item: item[1],
+                            reverse=True,
+                        )
+                    )
+                )
         df = pd.json_normalize(list(users.values()))
+
+        if hashtags:
+            df["countries"] = df["countries"].apply(lambda x: ",".join(map(str, x)))
+
         df = df.assign(
             changes=df["nodes.create"]
             + df["nodes.modify"]
@@ -462,6 +592,18 @@ def main():
         if args.rows:
             df = df.head(args.rows)
         print(df)
+        if args.wild_tags:
+            # Get the column names of the DataFrame
+            cols = df.columns.tolist()
+            # Identify the column names that you want to move
+            cols_to_move = ["tags_create", "tags_modify", "tags_delete"]
+            # Remove the columns to move from the list of column names
+            cols = [col for col in cols if col not in cols_to_move]
+            # Add the columns to move to the end of the list of column names
+            cols = cols + cols_to_move
+            # Reindex the DataFrame with the new order of column names
+            df = df.reindex(columns=cols)
+
         start_date = in_local_timezone(start_date, args.timezone)
         end_date = in_local_timezone(end_date, args.timezone)
 
@@ -470,7 +612,19 @@ def main():
             fname = args.name
         if "image" in args.format:
             # Convert the DataFrame to an image
-            dfi.export(df, f"{fname}.png", max_cols=-1)
+            df_img = (
+                (df if args.rows <= 100 else df.head(100))
+                if args.rows
+                else df.head(100)
+            )  # 100 as max rows for image format
+            dfi.export(
+                df_img.drop(columns=["tags_create", "tags_modify", "tags_delete"])
+                if args.wild_tags
+                else df_img,
+                f"{fname}.png",
+                max_cols=-1,
+                max_rows=-1,
+            )
 
         if "json" in args.format:
             # with open(f"{out_file_name}.json") as file:
@@ -513,6 +667,7 @@ def main():
         print("Metadata Created")
 
     else:
+        print("No data Found")
         sys.exit()
 
     end_time = time.time()
